#-*- coding:utf-8 -*-

#  Pybik -- A 3 dimensional magic cube game.
#  Copyright (C) 2009  B. Clausius
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License,  or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


import random

import cube
from .debug import *

rand = random.Random()
if DEBUG_LEVEL >= 0:
    debug('rand.seed(123)')
    rand.seed(123)


def slice_to_coord(slice, size):
    assert 0 <= slice < size, (slice, size)
    return float(slice * 2 - size + 1)

def coord_to_slice(coord, size):
    assert -size < coord < size
    return int(coord+size-1)//2

# This class is currently used to store the initial state of the cube
# and to provide the cube to plugin scripts.
#TODO: this class should replace the flubrd script
class CubeState (object):
    flubrd = 'FLUBRD'
    
    def __init__(self, dimension=None):
        self.type = 'Cube'
        if dimension is None:
            # Information is stored in module cube 
            self._dimension = None
            self._faces = None
            self._blocks = None
        else:
            assert 0 < dimension <= 10
            # solved cube
            self.set_solved(dimension=dimension)
    
    @property
    def dimension(self):
        if self._dimension is None:
            return cube.get_cube_dimension()
        return self._dimension
    @property
    def faces(self):
        if self._faces is None:
            return cube.make_faces_cube()
        return self._faces
    @property
    def blocks(self):
        if self._blocks is None:
            return cube.make_block_cube()
        return self._blocks
    
    def set_solved(self, dimension=None):
        assert self.type == 'Cube'
        if dimension:
            assert 0 < dimension <= 10
            self._dimension = dimension
        self._faces = 1
        self._blocks = 1  # 1 means solved cube
    
    @staticmethod
    def from_current():
        cubestate = CubeState()
        cubestate.get_current()
        return cubestate
    def get_current(self):
        self._dimension = cube.get_cube_dimension()
        self._faces = cube.make_faces_cube()
        self._blocks = cube.make_block_cube()
    
    def get_color(self, face, x, y):
        return self._faces[face][x + y*self._dimension]
        
    def dump(self):
        for i, f in enumerate(self._faces):
            debug('Face %d:'%i, str(f))
        debug()
        #for i, b in enumerate(self._blocks):
        #    debug('Block %d:'%i, str(b))
    
    #TODO this is the old interface to the cube, remove it
    @deprecated
    def __getitem__(self, i):
        if i == 0:
            return (1,3,self.dimension,self.dimension,self.dimension)
        if i == 1:
            return self.faces
        raise IndexError()
    
    #TODO: should move to cube.py
    rotation_matrices = [
        [[ 1,0,0,0],  [0, 1,0,0],  [0,0, 1,0]], # 0 empty
        [[ 1,0,0,0],  [0,-1,0,0],  [0,0,-1,0]], # 1 ff lluu uull
        [[-1,0,0,0],  [0, 1,0,0],  [0,0,-1,0]], # 2 ll ffuu uuff
        [[-1,0,0,0],  [0,-1,0,0],  [0,0, 1,0]], # 3 uu ffll llff
        
        [[ 1,0,0,0],  [0,0, 1,0],  [0,-1,0,0]], # 4 f
        [[ 1,0,0,0],  [0,0,-1,0],  [0, 1,0,0]], # 5 f'
        [[-1,0,0,0],  [0,0, 1,0],  [0, 1,0,0]], # 6 fuu uuf' f'll llf
        [[-1,0,0,0],  [0,0,-1,0],  [0,-1,0,0]], # 7 fll llf' f'uu uuf
        
        [[0, 1,0,0],  [ 1,0,0,0],  [0,0,-1,0]], # 8 ffu ull llu' u'ff
        [[0, 1,0,0],  [-1,0,0,0],  [0,0, 1,0]], # 9 u
        [[0,-1,0,0],  [ 1,0,0,0],  [0,0, 1,0]], #10 u'
        [[0,-1,0,0],  [-1,0,0,0],  [0,0,-1,0]], #11 ffu' u'll llu uff
        
        [[0, 1,0,0],  [0,0, 1,0],  [ 1,0,0,0]], #12 fu ul lf
        [[0, 1,0,0],  [0,0,-1,0],  [-1,0,0,0]], #13 f'u ul' l'f'
        [[0,-1,0,0],  [0,0, 1,0],  [-1,0,0,0]], #14 fu' u'l' l'f
        [[0,-1,0,0],  [0,0,-1,0],  [ 1,0,0,0]], #15 f'u' u'l lf'
        
        [[0,0, 1,0],  [ 1,0,0,0],  [0, 1,0,0]], #16 f'l' l'u' u'f'
        [[0,0, 1,0],  [-1,0,0,0],  [0,-1,0,0]], #17 uf fl' l'u
        [[0,0,-1,0],  [ 1,0,0,0],  [0,-1,0,0]], #18 fl lu' u'f
        [[0,0,-1,0],  [-1,0,0,0],  [0, 1,0,0]], #19 lu uf' f'l
        
        [[0,0, 1,0],  [0, 1,0,0],  [-1,0,0,0]], #20 l'
        [[0,0, 1,0],  [0,-1,0,0],  [ 1,0,0,0]], #21 ffl' l'uu uul lff
        [[0,0,-1,0],  [0, 1,0,0],  [ 1,0,0,0]], #22 l
        [[0,0,-1,0],  [0,-1,0,0],  [-1,0,0,0]], #23 ffl luu uul' l'ff
    ]
    
    rotation_symbolic = [
        "",     # 0 empty
        "ff",   # 1 lluu uull
        "ll",   # 2 ffuu uuff
        "uu",   # 3 ffll llff
        
        "f",    # 4
        "f'",   # 5
        "fuu",  # 6 uuf' f'll llf
        "fll",  # 7 llf' f'uu uuf
        
        "ffu",  # 8 ull llu' u'ff
        "u",    # 9 
        "u'",   #10 
        "ffu'", #11 u'll llu uff
        
        "fu",   #12 ul lf
        "f'u",  #13 ul' l'f'
        "fu'",  #14 u'l' l'f
        "f'u'", #15 u'l lf'
        
        "f'l'", #16 l'u' u'f'
        "uf",   #17 fl' l'u
        "fl",   #18 lu' u'f
        "lu",   #19 uf' f'l
        
        "l'",   #20 
        "ffl'", #21 l'uu uul lff
        "l",    #22 
        "ffl",  #23 luu uul' l'ff
    ]
    
    def format_block_compact(self):
        if self._blocks == 1:
            return '%s %d identity:' % (self.type,self.dimension)
        assert self._blocks
        res = '%s %d blocks_compact:' % (self.type, self.dimension)
        # every block is stored as 4-tuple (x,y,z,rot),
        # where rot is an index to a list of all (24) rotation matrices.
        for block in self._blocks:
            rot = self.rotation_matrices.index(block[:-1])
            x,y,z = [coord_to_slice(i, self.dimension) for i in block[3][:3]]
            res += ' %d,%d,%d,%s' % (x,y,z,self.rotation_symbolic[rot])
        return res
    
    def parse_block(self, code):
        tdf, blocks = code.split(':',1)
        t,d,f = tdf.split(' ',2)
        assert t == 'Cube'
        self._dimension = int(d)
        if f == 'identity':
            self._blocks = 1
            return
        assert f == 'blocks_compact'
        blocks = blocks.strip().split(' ')
        assert len(blocks) == self._dimension**3
        for i,block in enumerate(blocks):
            block = block.strip().split(',')
            assert len(block) == 4
            rot = self.rotation_symbolic.index(block[3])
            for j in range(3):
                block[j] = slice_to_coord(int(block[j]), self._dimension)
            block[3] = 1
            blocks[i] = self.rotation_matrices[rot] + [block]
        self._blocks = blocks
        
    def apply_current(self):
        if self._dimension:
            cube.set_cube_dimension(self._dimension)
        if self._blocks and self._blocks != 1:
            cube.cube_set_blocks(self._blocks)
        
    def randomize_and_apply(self):
        dimension = self._dimension
        if dimension:
            cube.set_cube_dimension(dimension)
        #debug('random moves:', 20*dimension)
        for i in xrange(20*dimension):
            axis = rand.randrange(3)
            dir = rand.randrange(2)
            if dimension>1 and dimension & 1:
                slice_ = rand.randrange(dimension - 1)
                if slice_ >= dimension // 2:
                    slice_ += 1
            else:
                slice_ = rand.randrange(dimension)
            slice_coord = slice_to_coord(slice_, dimension)
            cube.rotate_slice_4(axis, dir, slice_coord)
        self.get_current()
    def rotate_slice(self, move_data):
        cube.rotate_slice_4(move_data.axis, move_data.dir, move_data.slice)
        

